from .. import VEXObject

# IRConst heirarchy
class IRConst(VEXObject):
    type = None

    def __init__(self, c_expr):
        VEXObject.__init__(self)
        self.tag = ints_to_enums[c_expr.tag]

    def pp(self):
        print self.__str__()

    @property
    def size(self):
        return type_sizes[self.type]

    @staticmethod
    def _translate(c_expr):
        if c_expr[0] == ffi.NULL:
            return None

        tag = c_expr.tag

        try:
            return tag_to_class[tag](c_expr)
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRExprTag %s\n' % ints_to_enums[tag])

class U1(IRConst):
    type = 'Ity_I1'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.U1

    def __str__(self):
        return "%d" % self.value

class U8(IRConst):
    type = 'Ity_I8'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.U8

    def __str__(self):
        return "0x%02x" % self.value

class U16(IRConst):
    type = 'Ity_I16'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.U16

    def __str__(self):
        return "0x%04x" % self.value

class U32(IRConst):
    type = 'Ity_I32'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.U32

    def __str__(self):
        return "0x%08x" % self.value

class U64(IRConst):
    type = 'Ity_I64'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.U64

    def __str__(self):
        return "0x%016x" % self.value

class F32(IRConst):
    type = 'Ity_F32'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.F32

    def __str__(self):
        return "%f" % self.value

class F32i(IRConst):
    type = 'Ity_F32'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.F32

    def __str__(self):
        return "%f" % self.value

class F64(IRConst):
    type = 'Ity_F64'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.F64

    def __str__(self):
        return "%f" % self.value

class F64i(IRConst):
    type = 'Ity_F64'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.F64

    def __str__(self):
        return "%f" % self.value

class V128(IRConst):
    type = 'Ity_V128'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.V128

    def __str__(self):
        return "%x" % self.value

class V256(IRConst):
    type = 'Ity_V256'

    def __init__(self, c_expr):
        IRConst.__init__(self, c_expr)
        self.value = c_expr.Ico.V256

    def __str__(self):
        return "%x" % self.value

from .. import ints_to_enums, enums_to_ints, PyVEXError, ffi, type_sizes

tag_to_class = {
    enums_to_ints['Ico_U1']: U1,
    enums_to_ints['Ico_U8']: U8,
    enums_to_ints['Ico_U16']: U16,
    enums_to_ints['Ico_U32']: U32,
    enums_to_ints['Ico_U64']: U64,
    enums_to_ints['Ico_F32']: F32,
    enums_to_ints['Ico_F32i']: F32i,
    enums_to_ints['Ico_F64']: F64,
    enums_to_ints['Ico_F64i']: F64i,
    enums_to_ints['Ico_V128']: V128,
    enums_to_ints['Ico_V256']: V256,
}



from .. import VEXObject

# IRStmt heirarchy
class IRStmt(VEXObject):
	def __init__(self, c_stmt, irsb):
		VEXObject.__init__(self)
		self.arch = irsb.arch
		#self.c_stmt = c_stmt
		self.tag = ints_to_enums[c_stmt.tag]

	def pp(self):
		print self.__str__()

	@property
	def expressions(self):
		expressions = [ ]
		for _,v in self.__dict__.iteritems():
			if isinstance(v, IRExpr):
				expressions.append(v)
				expressions.extend(v.child_expressions)
		return expressions

	@property
	def constants(self):
		return sum((e.constants for e in self.expressions), [ ])

	@staticmethod
	def _translate(c_stmt, irsb):
		if c_stmt[0] == ffi.NULL:
			return None

		tag = c_stmt.tag
		try:
			stmt_class = _tag_to_class[tag]
		except KeyError:
			raise PyVEXError('Unknown/unsupported IRStmtTag %s\n' % ints_to_enums[tag])
		return stmt_class(c_stmt, irsb)

class NoOp(IRStmt):
	def __init__(self, c_stmt, irsb): #pylint:disable=unused-argument
		IRStmt.__init__(self, c_stmt, irsb)

	def __str__(self):
		return "IR-NoOp"

class IMark(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.addr = c_stmt.Ist.IMark.addr
		self.len = c_stmt.Ist.IMark.len
		self.delta = c_stmt.Ist.IMark.delta

	def __str__(self):
		return "------ IMark(0x%x, %d, %d) ------" % (self.addr, self.len, self.delta)

class AbiHint(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.base = IRExpr._translate(c_stmt.Ist.AbiHint.base, irsb)
		self.len = c_stmt.Ist.AbiHint.len
		self.nia = IRExpr._translate(c_stmt.Ist.AbiHint.nia, irsb)

	def __str__(self):
		return "====== AbiHint(0x%s, %d, %s) ======" % (self.base, self.len, self.nia)

class Put(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.data = IRExpr._translate(c_stmt.Ist.Put.data, irsb)
		self.offset = c_stmt.Ist.Put.offset

	def __str__(self):
		return "PUT(%s) = %s" % (self.arch.translate_register_name(self.offset), self.data)

class PutI(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.descr = IRRegArray(c_stmt.Ist.PutI.details.descr)
		self.ix = IRExpr._translate(c_stmt.Ist.PutI.details.ix, irsb)
		self.data = IRExpr._translate(c_stmt.Ist.PutI.details.data, irsb)
		self.bias = c_stmt.Ist.PutI.details.bias

	def __str__(self):
		return "PutI(%s)[%s,%d] = %s" % (self.descr, self.ix, self.bias, self.data)

class WrTmp(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.data = IRExpr._translate(c_stmt.Ist.WrTmp.data, irsb)
		self.tmp = c_stmt.Ist.WrTmp.tmp

	def __str__(self):
		return "t%d = %s" % (self.tmp, self.data)

class Store(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.addr = IRExpr._translate(c_stmt.Ist.Store.addr, irsb)
		self.data = IRExpr._translate(c_stmt.Ist.Store.data, irsb)
		self.end = ints_to_enums[c_stmt.Ist.Store.end]

	@property
	def endness(self):
		return self.end

	def __str__(self):
		return "ST%s(%s) = %s" % (self.endness[-2:].lower(), self.addr, self.data)

class CAS(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.addr = IRExpr._translate(c_stmt.Ist.CAS.details.addr, irsb)
		self.dataLo = IRExpr._translate(c_stmt.Ist.CAS.details.dataLo, irsb)
		self.dataHi = IRExpr._translate(c_stmt.Ist.CAS.details.dataHi, irsb)
		self.expdLo = IRExpr._translate(c_stmt.Ist.CAS.details.expdLo, irsb)
		self.expdHi = IRExpr._translate(c_stmt.Ist.CAS.details.expdHi, irsb)
		self.oldLo = c_stmt.Ist.CAS.details.oldLo
		self.oldHi = c_stmt.Ist.CAS.details.oldHi
		self.end = ints_to_enums[c_stmt.Ist.CAS.details.end]

	@property
	def endness(self):
		return self.end

	def __str__(self):
		return "t(%s,%s) = CAS%s(%s :: (%s,%s)->(%s,%s))" % (self.oldLo, self.oldHi, self.end[-2:].lower(), self.addr, self.expdLo, self.expdHi, self.dataLo, self.dataHi)

class LLSC(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.addr = IRExpr._translate(c_stmt.Ist.LLSC.addr, irsb)
		self.storedata = IRExpr._translate(c_stmt.Ist.LLSC.storedata, irsb)
		self.result = c_stmt.Ist.LLSC.result
		self.end = ints_to_enums[c_stmt.Ist.LLSC.end]

	@property
	def endness(self):
		return self.end

	def __str__(self):
		if self.storedata is None:
			return "t%d = LD%s-Linked(%s)" % (self.result, self.end[-2:].lower(), self.addr)
		else:
			return "t%d = ( ST%s-Cond(%s) = %s )" % (self.result, self.end[-2:].lower(), self.addr, self.storedata)

class MBE(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.event = ints_to_enums[c_stmt.Ist.MBE.event]

	def __str__(self):
		return "MBusEvent-" + self.event

class Dirty(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.cee = IRCallee(c_stmt.Ist.Dirty.details.cee)
		self.guard = IRExpr._translate(c_stmt.Ist.Dirty.details.guard, irsb)
		self.tmp = c_stmt.Ist.Dirty.details.tmp
		self.mFx = ints_to_enums[c_stmt.Ist.Dirty.details.mFx]
		self.mAddr = IRExpr._translate(c_stmt.Ist.Dirty.details.mAddr, irsb)
		self.mSize = c_stmt.Ist.Dirty.details.mSize
		self.nFxState = c_stmt.Ist.Dirty.details.nFxState

		self.args = [ ]
		for i in range(20):
			a = c_stmt.Ist.Dirty.details.args[i]
			if a == ffi.NULL:
				break

			self.args.append(IRExpr._translate(a, irsb))
		self.args = tuple(self.args)

	def __str__(self):
		return "t%s = DIRTY %s %s ::: %s(%s)" % (self.tmp, self.guard, "TODO(effects)", self.cee, ','.join(str(a) for a in self.args))

	@property
	def child_expressions(self):
		expressions = sum((a.child_expressions for a in self.args), [ ])
		expressions.extend(self.args)
		expressions.append(self.guard)
		expressions.extend(self.guard.child_expressions)
		return expressions

class Exit(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)
		self.guard = IRExpr._translate(c_stmt.Ist.Exit.guard, irsb)
		self.dst = IRConst._translate(c_stmt.Ist.Exit.dst)
		self.offsIP = c_stmt.Ist.Exit.offsIP
		self.jk = ints_to_enums[c_stmt.Ist.Exit.jk]

	@property
	def jumpkind(self):
		return self.jk

	def __str__(self):
		return "if (%s) { PUT(%s) = %s; %s }" % (self.guard, self.arch.translate_register_name(self.offsIP), hex(self.dst.value), self.jumpkind)

        @property
        def child_expressions(self):
                return [self.guard, self.dst] + self.guard.child_expressions

class LoadG(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.addr = IRExpr._translate(c_stmt.Ist.LoadG.details.addr, irsb)
		self.alt = IRExpr._translate(c_stmt.Ist.LoadG.details.alt, irsb)
		self.guard = IRExpr._translate(c_stmt.Ist.LoadG.details.guard, irsb)
		self.dst = c_stmt.Ist.LoadG.details.dst
		self.cvt = ints_to_enums[c_stmt.Ist.LoadG.details.cvt]

		self.end = ints_to_enums[c_stmt.Ist.LoadG.details.end]

		type_in = ffi.new('IRType *')
		type_out = ffi.new('IRType *')
		pvc.typeOfIRLoadGOp(c_stmt.Ist.LoadG.details.cvt, type_out, type_in)
		type_in = ffi.cast('int *', type_in)[0]
		type_out = ffi.cast('int *', type_out)[0]
		self.cvt_types = (ints_to_enums[type_in], ints_to_enums[type_out])

	@property
	def endness(self):
		return self.end

	def __str__(self):
		return "t%d = if (%s) %s(LD%s(%s)) else %s" % (self.dst, self.guard, self.cvt, self.end[-2:].lower(), self.addr, self.alt)

class StoreG(IRStmt):
	def __init__(self, c_stmt, irsb):
		IRStmt.__init__(self, c_stmt, irsb)

		self.addr = IRExpr._translate(c_stmt.Ist.StoreG.details.addr, irsb)
		self.data = IRExpr._translate(c_stmt.Ist.StoreG.details.data, irsb)
		self.guard = IRExpr._translate(c_stmt.Ist.StoreG.details.guard, irsb)
		self.end = ints_to_enums[c_stmt.Ist.StoreG.details.end]

	@property
	def endness(self):
		return self.end

	def __str__(self):
		return "if (%s) ST%s(%s) = %s" % (self.guard, self.end[-2:].lower(), self.addr, self.data)

from ..IRExpr import IRExpr
from ..IRConst import IRConst
from .. import IRRegArray, ints_to_enums, enums_to_ints, IRCallee, ffi, pvc, PyVEXError

_tag_to_class = {
	enums_to_ints['Ist_NoOp']: NoOp,
	enums_to_ints['Ist_IMark']: IMark,
	enums_to_ints['Ist_AbiHint']: AbiHint,
	enums_to_ints['Ist_Put']: Put,
	enums_to_ints['Ist_PutI']: PutI,
	enums_to_ints['Ist_WrTmp']: WrTmp,
	enums_to_ints['Ist_Store']: Store,
	enums_to_ints['Ist_LoadG']: LoadG,
	enums_to_ints['Ist_StoreG']: StoreG,
	enums_to_ints['Ist_CAS']: CAS,
	enums_to_ints['Ist_LLSC']: LLSC,
	enums_to_ints['Ist_Dirty']: Dirty,
	enums_to_ints['Ist_MBE']: MBE,
	enums_to_ints['Ist_Exit']: Exit,
}



import os
import urllib2
import subprocess
import sys
from distutils.errors import LibError
from distutils.core import setup
from distutils.command.build import build as _build

if sys.platform == 'darwin':
    library_file = "pyvex_static.dylib"
else:
    library_file = "pyvex_static.so"


VEX_LIB_NAME = "vex" # can also be vex-amd64-linux
VEX_PATH = "vex"
if not os.path.exists(VEX_PATH):
    VEX_URL = 'https://github.com/angr/vex/archive/master.tar.gz'
    with open('master.tar.gz', 'w') as v:
        v.write(urllib2.urlopen(VEX_URL).read())
    if subprocess.call(['tar', 'xzf', 'master.tar.gz']) != 0:
        raise LibError("Unable to retrieve libVEX.")
    VEX_PATH='./vex-master'

def _build_vex():
    if subprocess.call(['make'], cwd=VEX_PATH) != 0:
        raise LibError("Unable to build libVEX.")

def _build_pyvex():
    e = os.environ.copy()
    e['VEX_PATH'] = '../' + VEX_PATH
    if subprocess.call(['make'], cwd='pyvex_c', env=e) != 0:
        raise LibError("Unable to build pyvex-static.")

def _build_ffi():
    if subprocess.call(['python', 'make_ffi.py', os.path.join(VEX_PATH,'pub')]) != 0:
        raise LibError("Unable to generate cffi file.")

class build(_build):
    def run(self):
        self.execute(_build_vex, (), msg="Building libVEX")
        self.execute(_build_pyvex, (), msg="Building pyvex-static")
        self.execute(_build_ffi, (), msg="Creating CFFI defs file")
        _build.run(self)
cmdclass = { 'build': build }

try:
    from setuptools.command.develop import develop as _develop
    class develop(_develop):
        def run(self):
            self.execute(_build_vex, (), msg="Building libVEX")
            self.execute(_build_pyvex, (), msg="Building pyvex-static")
            self.execute(_build_ffi, (), msg="Creating CFFI defs file")
            _develop.run(self)
    cmdclass['develop'] = develop
except ImportError:
    print "Proper 'develop' support unavailable."

setup(
    name="pyvex", version='4.6.1.4', description="A Python interface to libVEX and VEX IR.",
    packages=['pyvex', 'pyvex.IRConst', 'pyvex.IRExpr', 'pyvex.IRStmt'],
    data_files=[
        ('lib', (os.path.join('pyvex_c', library_file),),),
    ],
    cmdclass=cmdclass,
    install_requires=[ 'pycparser', 'cffi>=1.0.3', 'archinfo' ]
)
